/* module.c: Sony Gesture Sensor Driver.
 *
 * Copyright 2016 Sony Corporation.
 * Author: Sony Corporation.
 *
 * This package is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * THIS PACKAGE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
 
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/miscdevice.h>
#include <linux/input.h>
#include <linux/dma-mapping.h>
#include <linux/crc32.h>
#include <linux/earlysuspend.h>
#include <asm/uaccess.h>
#include <mach/mt_gpio.h>
#include <mach/eint.h>
#include <icx_key.h>
#include "regs.h"

#if	defined(CONFIG_TARGET_MODEL_GUPPI)

	#include "library_guppi.h"

#elif 	defined(CONFIG_TARGET_MODEL_ANKOU)

	#if	defined(CONFIG_TARGET_ANKOU_BB)

		#include "library_guppi.h"

	#else
	
		#include "library_ankou.h"
		
	#endif
	
#elif	defined(CONFIG_TARGET_MODEL_VURI)

	#include "library_vuri.h"

#endif

#include "mgc3x30.h"

#define DRV_NAME					"mgc3x30"

#if 0
	#define DBGTRACE(fmt, ...)		                   printk("\e[32m[" DRV_NAME "] "         fmt "\e[0m\n", ## __VA_ARGS__)
	#define ERRTRACE(fmt, ...)		                   printk("\e[31m[" DRV_NAME "] "         fmt "\e[0m\n", ## __VA_ARGS__) ;  error_line = __LINE__
#else
	#define DBGTRACE(fmt, ...)		if( enable_log ) { printk(      "[" DRV_NAME "] "         fmt      "\n", ## __VA_ARGS__) ; }
	#define ERRTRACE(fmt, ...)		                   printk(      "[" DRV_NAME "] error - " fmt      "\n", ## __VA_ARGS__) ;  error_line = __LINE__
#endif

#define FIRMWARE_RETRY				3

	enum {
		BUFF_SIZE					=	132,						/* DMA バッファサイズ                     */
		I2C_SPEED					=	300,						/* I2C 通信速度                           */
		RESET_HOLD					=	30,							/* リセットホールド時間                   */
		RESET_RELEASE				=	50,							/* リセット解除待ち時間                   */
		BOOT_TIMEOUT				=	200,						/* 最初のバージョン受信のタイムアウト時間 */
		GETSYSTEMSTATUS_RETRY		=	10,							/* システムステータス取得のリトライ回数   */
		AIRWHEEL_CHATTER			=	5							/* エアホイールチャタリング回数           */
	} ;

	/* 状態 */
	enum {
		STATE_UNINIT,												/* 未初期化   */
		STATE_RUN,													/* 実行       */
		STATE_STOP,													/* 停止       */
		STATE_SUSPEND,												/* サスペンド */
		STATE_READ,													/* 読み込み   */
		STATE_WAIT,													/* 待ち       */
	} ;
	/* 実行フラグ */
	enum {
		FLAG_CHECKVERSION			=	1 << 0,						/* アップデート時にバージョンチェック */
	};
	/* ライブラリヘッダ */
	struct libheader {
		char						id[4] ;							/* 識別子（"GESL"） */
		u32							size ;							/* ライブラリサイズ */
	} ;

	/* プライベートデータ */
	struct priv {
		struct i2c_client			*p_client ;						/* I2C クライアント           */
		struct miscdevice			misc ;							/* MISC デバイス              */
		struct input_dev			*p_input ;						/* 入力デバイス               */
		wait_queue_head_t			waitQueueHead ;					/* ウェイトキュー             */
		dma_addr_t					buff_addr ;						/* DMA バッファの物理アドレス */
		u8							*p_buff ;						/* DMA バッファ               */
		struct workqueue_struct		*p_workqueue ;					/* ワークキュー               */
		struct work_struct			work ;							/* ワーク                     */
		struct early_suspend		early_suspend ;					/* サスペンド                 */
		struct mgc3x30_config		*p_config ;						/* 構成                       */
		int							eint ;							/* 割り込み番号               */
		int							mclr ;							/* MCLR ポート                */
		int							eio0 ;							/* EIO0 ポート                */
		u8							init ;							/* 初期化完了フラグ           */
		u8							stopFlag ;						/* 停止フラグ                 */
		u8							eintFlag ;						/* EINT フラグ                */
		u8							airWheelOld ;					/* 以前のエアホイール値       */
		s8							airWheelCount ;					/* ホイールカウンタ           */
	} ;


static struct priv					*m_p_priv ;						/* プライベートデータ               */
int									i2c_speed=I2C_SPEED ;			/* I2C 通信速度                     */
int									state=STATE_UNINIT ;			/* 状態                             */
int									fw_update=0 ;					/* ファームウェアアップデートフラグ */
int									irq_count=0 ;					/* 割り込みリクエスト数             */
int									error_line=0 ;					/* エラー発生行                     */
int									enable_log=0 ;					/* ログ有効フラグ                   */
static uint32_t						session_id = 0xa5a5a5a5;


module_param(i2c_speed,  int, 0664) ;
module_param(enable_log, int, 0664) ;
module_param(state,      int, 0444) ;
module_param(fw_update,  int, 0444) ;
module_param(irq_count,  int, 0664) ;
module_param(error_line, int, 0664) ;


/* -------------------------------------------------------------------------------- */
/*   配列から符号なし 16bit の値を取得しデータアドレスを更新                        */
/*                                                                                  */
/*   IN : pp_data = データアドレスのアドレス                                        */
/*   OUT: 値                                                                        */
/* -------------------------------------------------------------------------------- */
static inline u16 aryToU16( u8 **pp_data )
{
	u8	*p_data ;


	p_data   = *pp_data ;
	*pp_data = &p_data[2] ;

	return (((u16 )p_data[1] << 8) |
	        ((u16 )p_data[0] << 0)) ;
}


/* -------------------------------------------------------------------------------- */
/*   配列から符号なし 32bit の値を取得しデータアドレスを更新                        */
/*                                                                                  */
/*   IN : pp_data = データアドレスのアドレス                                        */
/*   OUT: 値                                                                        */
/* -------------------------------------------------------------------------------- */
static inline u32 aryToU32( u8 **pp_data )
{
	u8	*p_data ;


	p_data   = *pp_data ;
	*pp_data = &p_data[4] ;

	return (((u32 )p_data[3] << 24) |
	        ((u32 )p_data[2] << 16) |
	        ((u32 )p_data[1] <<  8) |
	        ((u32 )p_data[0] <<  0)) ;
}


/* -------------------------------------------------------------------------------- */
/*   EINT フラグを取得しクリア                                                      */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*   OUT: EINT フラグ状態                                                           */
/* -------------------------------------------------------------------------------- */
static int getEintFlag( struct priv *p_priv )
{
	u8	eintFlag ;


	eintFlag         = p_priv->eintFlag ;
	p_priv->eintFlag = 0 ;

	return eintFlag ;
}


/* -------------------------------------------------------------------------------- */
/*   ファームウェアバージョン情報を表示                                             */
/*                                                                                  */
/*   IN : p_fwVersionInfo = ファームウェアバージョン情報格納アドレス                */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void showFwVersionInfo( struct fw_version_info *p_fwVersionInfo )
{
	char	fwVersion[sizeof( p_fwVersionInfo->fwVersion )+1], *p_fwVersion, *p_next ;
	int		count ;


	printk("fw_version_info:\n") ;
	printk("  FwValid              = 0x%02x\n",   p_fwVersionInfo->fwValid) ;
	printk("  HwRev                = %d.%d\n",    p_fwVersionInfo->hwRev[0], p_fwVersionInfo->hwRev[1]) ;
	printk("  ParameterStartAddr   = 0x%04x\n",   128*p_fwVersionInfo->parameterStartAddr) ;
	printk("  LibraryLoaderVersion = %d.%d.%d\n", p_fwVersionInfo->libraryLoaderVersion[0],
	                                              p_fwVersionInfo->libraryLoaderVersion[1],
	                                              p_fwVersionInfo->libraryLoaderVersion[2]) ;
	printk("  FwStartAddr          = 0x%04x\n",   128*p_fwVersionInfo->fwStartAddr) ;

	printk("  FwVersion            = ") ;

	memcpy(&fwVersion[0], &p_fwVersionInfo->fwVersion[0], sizeof( p_fwVersionInfo->fwVersion )) ;
	fwVersion[sizeof( p_fwVersionInfo->fwVersion )] = '\0' ;

	p_fwVersion = &fwVersion[0] ;
	count       = 0 ;

	while( 1 ) {
		p_next = strchr(&p_fwVersion[0], ';') ;
		if( p_next != NULL ) {
			*p_next++ = '\0' ;
		}

		if( p_fwVersion[0] != '\0' ) {
			if( count++ ) {
				printk("                         ") ;
			}
			printk("%s\n", &p_fwVersion[0]) ;
		}

		if( p_next == NULL ) {
			break ;
		}

		p_fwVersion = &p_next[0] ;
	}

	if( !count ) {
		printk("none\n") ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   データ読み込み                                                                 */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*        p_buff = 読み込みバッファ                                                 */
/*        size   = 読み込みバッファサイズ                                           */
/*        retry  = リトライ回数                                                     */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/*        -ETIMEDOUT                                                                */
/*        -ECANCELED                                                                */
/* -------------------------------------------------------------------------------- */
static int readData(struct priv *p_priv, void *p_buff, u8 size, uint timeout)
{
	struct i2c_msg	msg ;
	int				result ;


	if( timeout ) {
		result = wait_event_timeout(p_priv->waitQueueHead, getEintFlag( p_priv ), msecs_to_jiffies( timeout )) ;
		if( !result ) {
			return -ETIMEDOUT ;
		}
	}

	mt_eint_mask( p_priv->eint ) ;
	mt_set_gpio_dir(p_priv->eio0, GPIO_DIR_OUT) ;

	memset(&msg, 0, sizeof( msg )) ;
	msg.addr     = p_priv->p_client->addr ;
	msg.flags    = I2C_M_RD ;
	msg.len      = size ;
	msg.buf      = (void *)p_priv->buff_addr ;
	msg.timing   = i2c_speed ;
	msg.ext_flag = I2C_DMA_FLAG ;
	result       = i2c_transfer(p_priv->p_client->adapter, &msg, 1) ;

	mt_eint_unmask( p_priv->eint ) ;
	mt_set_gpio_dir(p_priv->eio0, GPIO_DIR_IN) ;

	if( result != 1 ) {
		return -EIO ;
	}

	/* データコピー */
	memcpy(p_buff, &p_priv->p_buff[0], size) ;

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   データ書き込み                                                                 */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*        p_buff = 書き込みバッファ                                                 */
/*        size   = 書き込みバッファサイズ                                           */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/* -------------------------------------------------------------------------------- */
static int writeData(struct priv *p_priv, const void *p_buff, u16 size)
{
	struct i2c_msg	msg ;
	int				result ;

	/* データコピー */
	memcpy(&p_priv->p_buff[0], p_buff, size) ;

	memset(&msg, 0, sizeof( msg )) ;
	msg.addr     = p_priv->p_client->addr ;
	msg.len      = size ;
	msg.buf      = (void *)p_priv->buff_addr ;
	msg.timing   = i2c_speed ;
	msg.ext_flag = I2C_DMA_FLAG ;
	result       = i2c_transfer(p_priv->p_client->adapter, &msg, 1) ;
	if( result != 1 ) {
		return -EIO ;
	}

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   ファームウェアバージョン情報を取得                                             */
/*                                                                                  */
/*   IN : p_priv          = プライベートデータ                                      */
/*        p_fwVersionInfo = ファームウェアバージョン情報格納アドレス                */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/*        -ETIMEDOUT                                                                */
/*        -ENODEV                                                                   */
/* -------------------------------------------------------------------------------- */
static int getFwVersionInfo(struct priv *p_priv, struct fw_version_info *p_fwVersionInfo)
{
	int		result ;

	/* ファームウェアバージョン取得 */
	result = readData(p_priv, p_fwVersionInfo, sizeof( struct fw_version_info ), 500) ;
	if( result ) {
		return result ;
	}

	/* GestIC ライブラリ 異常 */
	if( p_fwVersionInfo->fwValid != ValidFW ) {
		return -ENODEV ;
	}

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   システムステータスを取得                                                       */
/*                                                                                  */
/*   IN : p_priv         = プライベートデータ                                       */
/*        msgID          = メッセージ ID                                            */
/*        p_systemStatus = システムステータス格納アドレス                           */
/*   OUT: エラーコード                                                              */
/*        -EIO                                                                      */
/*        -ETIMEDOUT                                                                */
/* -------------------------------------------------------------------------------- */
static int getSystemStatus(struct priv *p_priv, u16 msgID, struct system_status *p_systemStatus)
{
	int		result, retry ;


	for(retry=0; retry<GETSYSTEMSTATUS_RETRY; retry++) {
		result = readData(p_priv, p_systemStatus, sizeof( struct system_status ), 100) ;
		if( result ) {
			return result ;
		}

		if( (p_systemStatus->header.ID==System_Status) && (p_systemStatus->msgID==msgID) ) {
			return le16_to_cpu( p_systemStatus->errorCode ) ;
		}
	}

	return -ETIMEDOUT ;
}

#if 0
/* -------------------------------------------------------------------------------- */
/*   ランタイムパラメータを設定                                                     */
/*                                                                                  */
/*   IN : p_priv    = プライベートデータ                                            */
/*        ID        = ランタイムパラメータ ID                                       */
/*        argument0 = 引数０                                                        */
/*        argument1 = 引数１                                                        */
/*   OUT: エラーコード                                                              */
/*        -EIO                                                                      */
/*        -ETIMEDOUT                                                                */
/* -------------------------------------------------------------------------------- */
static int setRuntimeParameter(struct priv *p_priv, u16 ID, u32 argument0, u32 argument1)
{
	union {
		struct runtime_parameter	runtimeParameter ;
		struct system_status		systemStatus ;
	} msg ;
	int								result ;

	/* ランタイムパラメータ設定 */
	msg.runtimeParameter.header.size        = sizeof( struct runtime_parameter ) ;
	msg.runtimeParameter.header.flags       = 0 ;
	msg.runtimeParameter.header.seq         = 0 ;
	msg.runtimeParameter.header.ID          = Set_Runtime_Parameter ;
	msg.runtimeParameter.runtimeParameterID = ID ;
	msg.runtimeParameter.reserved           = 0 ;
	msg.runtimeParameter.argument[0]        = cpu_to_le32( argument0 ) ;
	msg.runtimeParameter.argument[1]        = cpu_to_le32( argument1 ) ;
	result                                  = writeData(p_priv, &msg.runtimeParameter, sizeof( struct runtime_parameter )) ;
	if( result ) {
		return result ;
	}

	return getSystemStatus(p_priv, Set_Runtime_Parameter, &msg.systemStatus) ;
}
#endif


/* -------------------------------------------------------------------------------- */
/*   ホスト割り込み処理                                                             */
/*                                                                                  */
/*   IN : なし                                                                      */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void host_isr( void )
{
	struct priv		*p_priv=m_p_priv ;


	DBGTRACE("%s(%d)", __func__, __LINE__) ;

	irq_count++ ;

	p_priv->eintFlag = 1 ;

	wake_up( &p_priv->waitQueueHead ) ;
}

static void sensorDataIsr( void )
{
	struct priv		*p_priv=m_p_priv ;


	DBGTRACE("%s(%d)", __func__, __LINE__) ;

	irq_count++ ;

	queue_work(p_priv->p_workqueue, &p_priv->work) ;
}


/* -------------------------------------------------------------------------- */
/*   ワークキュー処理                                                         */
/*                                                                            */
/*   IN : p_wok = プライベートデータ内のワークアドレス                        */
/*   OUT: なし                                                                */
/* -------------------------------------------------------------------------- */
static void	workqueue_proc( struct work_struct *p_work )
{
	struct priv				*p_priv=container_of(p_work, struct priv, work) ;
	struct mgc3x30_config	*p_config=p_priv->p_config ;
	union {
		struct sensor_data	data ;
		u8					buff[BUFF_SIZE] ;
	} sensor ;
	s8						diff ;
	u8						*p_data ;
	u16						dspStatus, airWheelInfo, x, y, z ;
	u32						gestureInfo, touchInfo ;
	int						result ;
	int						const *p_keyInfo ;


	while( !(mt_get_gpio_in( p_priv->eio0 )|p_priv->stopFlag) ) {

		state = STATE_READ ;

		/* センサデータ取得 */
		result = readData(p_priv, &sensor, p_config->read_size, 0) ;
		if( result ) {
			ERRTRACE("%s(%d)", __func__, __LINE__) ;
			continue ;
		}

		state = STATE_WAIT ;

		DBGTRACE("Sensor_Data_Output : 0x%02x, %d, 0x%04x", sensor.data.header.ID, sensor.data.header.size, sensor.data.dataOutputConfigMask) ;

		if( sensor.data.header.ID != 0x91 ) {
			continue ;
		}

		p_data = (u8 *)&sensor.data.data[0] ;

		/* DSPStatus */
		if( sensor.data.dataOutputConfigMask & DSPStatus ) {
			dspStatus = aryToU16( &p_data ) ;
			DBGTRACE("  dspStatus    = 0x%04x", dspStatus) ;
		}

		/* GestureInfo */
		if( sensor.data.dataOutputConfigMask & GestureInfo ) {
			gestureInfo = aryToU32( &p_data ) ;
			DBGTRACE("  gestureInfo  = 0x%08x", gestureInfo) ;
			p_keyInfo   = &p_config->p_gesture[0] ;
			if( p_keyInfo != NULL ) {
				do {
					if( (u8 )gestureInfo == p_keyInfo[0] ) {
						input_event(p_priv->p_input, EV_KEY, p_keyInfo[1], 1) ;
						input_sync( p_priv->p_input ) ;
						input_event(p_priv->p_input, EV_KEY, p_keyInfo[1], 0) ;
						input_sync( p_priv->p_input ) ;
						break ;
					}
					p_keyInfo += 2 ;
				} while( p_keyInfo[0] ) ;
			}
		}

		/* TouchInfo */
		if( sensor.data.dataOutputConfigMask & TouchInfo ) {
			touchInfo = aryToU32( &p_data ) ;
			DBGTRACE("  touchInfo    = 0x%08x", touchInfo) ;
		}

		/* AirWheelInfo */
		if( (sensor.data.dataOutputConfigMask&AirWheelInfo) && (sensor.data.systemInfo&AirWheelValid) ) {
			airWheelInfo = aryToU16( &p_data ) ;
			p_keyInfo    = &p_config->p_airwheel[0] ;

			DBGTRACE("  airWheelInfo = %d, %d", airWheelInfo, p_priv->airWheelCount) ;

			if( p_keyInfo != NULL ) {
				diff                = (u8 )airWheelInfo - p_priv->airWheelOld ;
				p_priv->airWheelOld = (u8 )airWheelInfo ;
				if( diff > 0 ) {
					if( ++p_priv->airWheelCount >  AIRWHEEL_CHATTER ) {
						p_priv->airWheelCount = 0 ;
						input_event(p_priv->p_input, EV_KEY, p_keyInfo[0], 1) ;
						input_sync( p_priv->p_input ) ;
						input_event(p_priv->p_input, EV_KEY, p_keyInfo[0], 0) ;
						input_sync( p_priv->p_input ) ;
					}
				} else
				if( diff < 0 ) {
					if( --p_priv->airWheelCount < -AIRWHEEL_CHATTER ) {
						p_priv->airWheelCount = 0 ;
						input_event(p_priv->p_input, EV_KEY, p_keyInfo[1], 1) ;
						input_sync( p_priv->p_input ) ;
						input_event(p_priv->p_input, EV_KEY, p_keyInfo[1], 0) ;
						input_sync( p_priv->p_input ) ;
					}
				}
			}
		} else {
			p_priv->airWheelCount = 0 ;
		}

		/* xyzPosition */
		if( (sensor.data.dataOutputConfigMask&xyzPosition) && (sensor.data.systemInfo&PositionValid) ) {
			x = aryToU16( &p_data ) ;
			y = aryToU16( &p_data ) ;
			z = aryToU16( &p_data ) ;
			DBGTRACE("  xyzPosition  = %d, %d, %d", x, y, z) ;
		}
	}
}


/* -------------------------------------------------------------------------------- */
/*   リセット                                                                       */
/*                                                                                  */
/*   IN : p_priv = プライベートデータ                                               */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void reset( struct priv *p_priv )
{
	mt_eint_mask( p_priv->eint ) ;

	mt_set_gpio_out(p_priv->mclr, GPIO_OUT_ZERO) ;

	mt_set_gpio_dir(p_priv->eio0, GPIO_DIR_IN) ;
	mt_set_gpio_out(p_priv->eio0, GPIO_OUT_ZERO) ;

	msleep( RESET_HOLD ) ;

	mt_set_gpio_out(p_priv->mclr, GPIO_OUT_ONE) ;

	msleep( RESET_RELEASE ) ;

	mt_eint_unmask( p_priv->eint ) ;
}


/* -------------------------------------------------------------------------------- */
/*   ファームウェア書き換え開始                                                     */
/*                                                                                  */
/*   IN : p_priv    = プライベートデータ                                            */
/*   OUT: エラーコード                                                              */
/*        -EIO                                                                      */
/*        -ETIMEDOUT                                                                */
/* -------------------------------------------------------------------------------- */
static int StartFirmwareUpdate(struct priv *p_priv, struct library *p_library)
{
	
	struct{
		struct update_start				updateStart;
		struct system_status			systemStatus;
	}msg;
	
	int									result ;
	
	memset(&msg, 0, sizeof( msg )) ;
	
	msg.updateStart.header.size			= sizeof( struct update_start );
	msg.updateStart.header.flags		= 0;
	msg.updateStart.header.seq			= 0;
	msg.updateStart.header.ID			= Fw_Update_Start;
	
	msg.updateStart.sessionId			= ++session_id;
	
	memcpy(msg.updateStart.iv, p_library->iv, sizeof(msg.updateStart.iv));

	msg.updateStart.updateFunction		= ProgramFlash;
	msg.updateStart.reserved			= 0;
	
	msg.updateStart.crc =~ crc32_le(0xffffffff, (const unsigned char *)&msg.updateStart.sessionId, sizeof(struct update_start) - 8);
	
	result                                  = writeData(p_priv, &msg.updateStart, sizeof( struct update_start)) ;

	if(result == 0)
	{
		result								= getSystemStatus(p_priv, Fw_Update_Start, &msg.systemStatus) ;
	}
	
	msleep(1);
	
	return result;

}
/* -------------------------------------------------------------------------------- */
/*   ファームウェアデータ送信                                                       */
/*                                                                                  */
/*   IN : p_priv    = プライベートデータ                                            */
/*   OUT: エラーコード                                                              */
/*        -EIO                                                                      */
/*        -ETIMEDOUT                                                                */
/* -------------------------------------------------------------------------------- */
static int SendFirmwareData(struct priv *p_priv, struct library *p_library)
{
	
	struct{
		struct update_block				updateBlock;
		struct system_status			systemStatus;
	}msg;
	
	int									result ;
	int									block_number;
	
	memset(&msg, 0, sizeof( msg )) ;
	
	msg.updateBlock.header.size      = sizeof( struct update_block );
	msg.updateBlock.header.flags     = 0;
	msg.updateBlock.header.seq       = 0;
	msg.updateBlock.header.ID        = Fw_Update_Block;
	msg.updateBlock.updateFunction   = ProgramFlash;
	
	for(block_number=0; block_number < p_library->record_count; block_number++)
	{
	
		msg.updateBlock.address          = p_library->data[block_number].address;
		msg.updateBlock.length           = p_library->data[block_number].length;
		
		memcpy(msg.updateBlock.payload, p_library->data[block_number].data, p_library->data[block_number].length);

		msg.updateBlock.crc =~ crc32_le(0xffffffff, (const unsigned char *)&msg.updateBlock.address, sizeof(struct update_block) - 8);
		
		result                                  = writeData(p_priv, &msg.updateBlock, sizeof( struct update_block));
		
		if( result != 0)
		{
			break;
		}
		
		result									= getSystemStatus(p_priv, Fw_Update_Block, &msg.systemStatus) ;
		
		if( result != 0)
		{
			break;
		}
		
	}
	
	msleep(1);
	
	return result;
}
/* -------------------------------------------------------------------------------- */
/*   ファームウェアアップデート完了                                                 */
/*                                                                                  */
/*   IN : p_priv    = プライベートデータ                                            */
/*   OUT: エラーコード                                                              */
/*        -EIO                                                                      */
/*        -ETIMEDOUT                                                                */
/* -------------------------------------------------------------------------------- */
static int CompleteFirmwareUpdate(struct priv *p_priv, u8 u_update, struct library *p_library)
{
	
	struct{
		struct update_completed			  updateCompleted;
		struct system_status			  systemStatus;
	}msg;
	
	int								result ;
	
	memset(&msg, 0, sizeof( msg )) ;
	
	msg.updateCompleted.header.size    = sizeof( struct update_completed );
	msg.updateCompleted.header.flags   = 0;
	msg.updateCompleted.header.seq     = 0;
	msg.updateCompleted.header.ID      = Fw_Update_Completed;
	msg.updateCompleted.sessionId      = session_id;
	msg.updateCompleted.updateFunction = u_update;
	
	memcpy(msg.updateCompleted.fwVersion, p_library->fw_version,
		sizeof(msg.updateCompleted.fwVersion));

	memset(msg.updateCompleted.reserved, 0, sizeof(msg.updateCompleted.reserved));
	
	msg.updateCompleted.crc =~ crc32_le(0xffffffff, (const unsigned char *)&msg.updateCompleted.sessionId, sizeof(struct update_completed) - 8);
	
	result                                  = writeData(p_priv, &msg.updateCompleted, sizeof( struct update_completed)) ;
	
	if(result == 0){
		result								= getSystemStatus(p_priv, Fw_Update_Completed, &msg.systemStatus) ;
	}
	
	msleep(1);
	
	return result;
}
/* -------------------------------------------------------------------------------- */
/*   ファームウェア書き換え実施                                                     */
/*                                                                                  */
/*   IN : p_priv    = プライベートデータ                                            */
/*   OUT: エラーコード                                                              */
/*        -EIO                                                                      */
/*        -ETIMEDOUT                                                                */
/* -------------------------------------------------------------------------------- */
static int mgc3x30_fw_update(struct priv *p_priv, struct library *p_library)
{
	int	retval;
	int	i;
	DBGTRACE("firmware update start\n");
	
	for (i = 0; i < FIRMWARE_RETRY; i++) 
	{
		// 書き込み開始コマンド送信
		retval = StartFirmwareUpdate(p_priv, p_library);
		if (retval) {
			reset( p_priv ) ;
			ERRTRACE("%s(%d)\n", __func__, __LINE__);
			continue;
		}
		
		// アップデートブロックコマンド送信
		retval = SendFirmwareData(p_priv, p_library);
		if (retval) {
			reset( p_priv ) ;
			ERRTRACE("%s(%d)\n", __func__, __LINE__);
			continue;
		}
		
		// 書き込み完了コマンドを送信
		retval = CompleteFirmwareUpdate(p_priv, ProgramFlash, p_library);
		if (retval) {
			reset( p_priv ) ;
			ERRTRACE("%s(%d)\n", __func__, __LINE__);
			continue;
		}
		// Restart 開始 
		retval = CompleteFirmwareUpdate(p_priv, Restart, p_library);
		if (retval) {
			reset( p_priv ) ;
			ERRTRACE("%s(%d)\n", __func__, __LINE__);
			continue;
		}
		/* 起動待ち */
		DBGTRACE("restarting\n");
		msleep( BOOT_TIMEOUT ) ;
		DBGTRACE("firmware update success\n");
		
		break;	// no error detected

	}
	if(retval !=0)
	{
		ERRTRACE("firmware update failed %s(%d)\n", __func__, __LINE__);
	}
	return retval;	// success
}
/* -------------------------------------------------------------------------------- */
/*   バージョン比較処理                                                             */
/*                                                                                  */
/*   IN : u_version    = バージョンデータ                                           */
/*   OUT: 0 バージョン一致                                                          */
/*        0以外 バージョン不一致                                                    */
/* -------------------------------------------------------------------------------- */
static int chkFWVersion(uint8_t *u_version)
{
	int  i=0;
	int  size=0;
	
	for(i=0; i < sizeof(Library.fw_version); i++){
		if(Library.fw_version[i] != 0x0){
				size++;
		}
		else{
			break;
		}
	}
	
	return( memcmp(u_version, Library.fw_version, size ));
}
/* -------------------------------------------------------------------------------- */
/*   実行                                                                           */
/*                                                                                  */
/*   IN : p_priv    = プライベートデータ                                            */
/*        p_library = ライブラリ（NULL 時は更新しない）                             */
/*        check     = ライブラリバージョンチェック                                  */
/*   OUT: 0                                                                         */
/*        -EIO                                                                      */
/*        -ETIMEDOUT                                                                */
/*        -ENODEV                                                                   */
/* -------------------------------------------------------------------------------- */
static int run(struct priv *p_priv, struct library *p_library, uint flag)
{
	struct fw_version_info	fwVersionInfo ;
	int						result, update ;

	/* 割り込み設定 */
	mt_eint_mask( p_priv->eint ) ;
	mt_eint_registration(p_priv->eint, EINTF_TRIGGER_FALLING, host_isr, 1) ;
	mt_eint_mask( p_priv->eint ) ;

	/* リセット */
	reset( p_priv ) ;

	do {
		/* ファームウェアバージョン取得 */
		result = getFwVersionInfo(p_priv, &fwVersionInfo) ;
		if( result ) {
			ERRTRACE("%s(%d)\n", __func__, __LINE__) ;

			if( p_library == NULL ) {
				break ;
			}
		}

		if( p_library != NULL ) {

			/* FWステータスとバージョンチェック */
			update = 1 ;
			if( !result && (flag&FLAG_CHECKVERSION) ) {
				if(( fwVersionInfo.fwValid == ValidFW ) && (chkFWVersion( fwVersionInfo.fwVersion ) == 0 ) )
				{
					update = 0 ;
				}
			}

			if( update )
			{
				fw_update = 1 ;

				// FW書き換え実施
				result = mgc3x30_fw_update(p_priv, p_library);
				if( result ) 
				{
					ERRTRACE("%s(%d)", __func__, __LINE__) ;
					break ;
				}

				/* リセット */
				reset( p_priv ) ;

				/* ファームウェアバージョン取得 */
				result = getFwVersionInfo(p_priv, &fwVersionInfo) ;
				if( result ) {
					ERRTRACE("%s(%d)\n", __func__, __LINE__) ;
					break ;
				}
			}
		}

		/* 起動待ち */
		msleep( BOOT_TIMEOUT ) ;

		/* ファームウェアバージョン表示 */
		showFwVersionInfo( &fwVersionInfo ) ;

		/* 割り込み変更 */
		mt_eint_mask( p_priv->eint ) ;
		mt_eint_registration(p_priv->eint, EINTF_TRIGGER_FALLING, sensorDataIsr, 1) ;
		mt_eint_unmask( p_priv->eint ) ;

		/* ワークキュー動作 */
		queue_work(p_priv->p_workqueue, &p_priv->work) ;

		state = STATE_RUN ;

		/* 正常終了 */
		return 0 ;
	} while( 0 ) ;

	mt_eint_mask( p_priv->eint ) ;

	state = STATE_STOP ;

	return result ;
}


/* -------------------------------------------------------------------------------- */
/*   停止                                                                           */
/*                                                                                  */
/*   IN : p_priv プライベートデータ                                                 */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void  stop( struct priv *p_priv )
{
	/* 割り込み解除 */
	mt_eint_mask( p_priv->eint ) ;
	mt_eint_registration(p_priv->eint, EINTF_TRIGGER_FALLING, NULL, 1) ;
	mt_eint_mask( p_priv->eint ) ;

	p_priv->stopFlag = 1 ;

	cancel_work_sync( &p_priv->work ) ;

	/* リセット状態 */
	mt_set_gpio_dir(p_priv->eio0, GPIO_DIR_IN) ;
	mt_set_gpio_out(p_priv->eio0, GPIO_OUT_ZERO) ;
	mt_set_gpio_out(p_priv->mclr, GPIO_OUT_ZERO) ;

	p_priv->stopFlag      = 0 ;
	p_priv->airWheelCount = 0 ;
}


/* -------------------------------------------------------------------------------- */
/*   ライブラリファイルを読み込む                                                   */
/*                                                                                  */
/*   IN : p_fileName = ファイル名                                                   */
/*   OUT: NULL      != ライブラリ（kfree）                                          */
/*        NULL      == 異常終了                                                     */
/* -------------------------------------------------------------------------------- */
static void *readLibrary( const char *p_fileName )
{
	mm_segment_t		fs ;
	struct file			*p_file ;
	struct libheader	libHeader ;
	loff_t				size, offset ;
	u8					*p_buff ;
	int					result ;


	fs = get_fs() ;

	set_fs( KERNEL_DS ) ;

	p_buff = NULL ;
	offset = 0 ;
	result = -1 ;

	do {
		p_file = filp_open(&p_fileName[0], O_RDONLY, 0) ;
		if( IS_ERR( p_file ) ) {
			break ;
		}

		/* ファイルサイズ取得 */
		size = vfs_llseek(p_file, 0, SEEK_END) ;
		vfs_llseek(p_file, 0, SEEK_SET) ;

		if( size <= sizeof( libHeader  ) ) {
			break ;
		}

		result = vfs_read(p_file, (char *)&libHeader, sizeof( libHeader ), &offset) ;
		if( result != sizeof( libHeader ) ) {
			result = -1 ;
			break ;
		}
		size -= sizeof( libHeader ) ;

		if( memcmp(&libHeader.id[0], "GESL", 4) || (size != libHeader.size) ) {
			break ;
		}

		p_buff = kmalloc(size, GFP_KERNEL) ;
		if( p_buff == NULL ) {
			break ;
		}

		result = vfs_read(p_file, &p_buff[0], size, &offset) ;
		if( result != size ) {
			result = -1 ;
			break ;
		}

		/* 正常終了 */
		result = 0 ;
	} while( 0 ) ;

	set_fs( fs ) ;

	if( result ) {
		if( !IS_ERR( p_file ) ) {
			filp_close(p_file, NULL) ;
		}

		if( p_buff != NULL ) {
			kfree( p_buff ) ;
			p_buff = NULL ;
		}
	}

	return p_buff ;
}


/* -------------------------------------------------------------------------------- */
/*   開く                                                                           */
/*                                                                                  */
/*   IN : p_inode = i ノード                                                        */
/*        p_file  = ファイル                                                        */
/*   OUT: 0                                                                         */
/* -------------------------------------------------------------------------------- */
static int mgc3x30_open(struct inode *p_inode, struct file *p_file)
{
	p_file->private_data = container_of(p_file->private_data, struct priv, misc) ;

	return 0 ;
}


/* -------------------------------------------------------------------------------- */
/*  書き込み                                                                        */
/*                                                                                  */
/*  IN : p_file = ファイル                                                          */
/*       p_buff = 書き込みバッファ                                                  */
/*       size   = 書き込みサイズ                                                    */
/*       p_pos  = 位置格納アドレス                                                  */
/*  OUT: 書き込みサイズ                                                             */
/* -------------------------------------------------------------------------------- */
static ssize_t mgc3x30_write(struct file *p_file, const char __user *p_buff, size_t size, loff_t *p_pos)
{
	struct priv		*p_priv=p_file->private_data ;
	struct library	*p_library ;
	char			fileName[128] ;
	int				result ;


	if( (size<1) || (size>=sizeof( fileName )) ) {
		return -EINVAL ;
	}

	if( copy_from_user(&fileName[0], &p_buff[0], size) ) {
		return -EFAULT ;
	}

	fileName[size] = '\0' ;
	if( fileName[size-1] == '\n' ) {
		fileName[size-1] = '\0' ;
	}

	p_library = readLibrary( &fileName[0] ) ;
	if( p_library == NULL ) {
		return -EIO ;
	}

	if( p_priv->init ) {
		p_priv->init = 0 ;
		stop( p_priv ) ;
	}

	result = run(p_priv, p_library, 0) ;

	kfree( p_library ) ;

	if( result ) {
		return result ;
	}

	p_priv->init = 1 ;

	return size ;
}


/* -------------------------------------------------------------------------------- */
/*   ioctl                                                                          */
/*                                                                                  */
/*   IN : p_file = ファイル                                                         */
/*        cmd    = コマンド                                                         */
/*        arg    = 引数                                                             */
/*   OUT: 0                                                                         */
/*        キーコード数                                                              */
/*        -EFAULT                                                                   */
/*        -EINVAL                                                                   */
/* -------------------------------------------------------------------------------- */
static long mgc3x30_ioctl(struct file *p_file, unsigned int cmd, unsigned long arg)
{
	struct priv		*p_priv=p_file->private_data ;


	switch( cmd ) {
		case ICX_KEY_IOCTL_CMD_GETCOUNT:
		{
			return p_priv->p_config->count ;
		}

		case ICX_KEY_IOCTL_CMD_GETINFO:
		{
			if( copy_to_user((void __user *)arg, &p_priv->p_config->p_key_code[0], sizeof( u32 )*p_priv->p_config->count) ) {
				return -EFAULT ;
			}
			break ;
		}

		default:
			return -EINVAL ;
	}

	return 0 ;
}


static const struct file_operations		mgc3x30_fops={
											.owner          = THIS_MODULE,
											.open           = mgc3x30_open,
											.write          = mgc3x30_write,
											.unlocked_ioctl = mgc3x30_ioctl,
											.compat_ioctl   = mgc3x30_ioctl,
										} ;


/* -------------------------------------------------------------------------------- */
/*   サスペンド処理                                                                 */
/*                                                                                  */
/*   IN : p_early_suspend = サスペンド                                              */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void mgc3x30_early_suspend( struct early_suspend *p_early_suspend )
{
	struct priv		*p_priv=container_of(p_early_suspend, struct priv, early_suspend) ;


	DBGTRACE("%s(%d)", __func__, __LINE__) ;

	if( p_priv->init ) {
		state = STATE_SUSPEND ;
		stop( p_priv ) ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   レジューム処理                                                                 */
/*                                                                                  */
/*   IN : p_early_suspend = サスペンド                                              */
/*   OUT: なし                                                                      */
/* -------------------------------------------------------------------------------- */
static void mgc3x30_early_resume( struct early_suspend *p_early_suspend )
{
	struct priv		*p_priv=container_of(p_early_suspend, struct priv, early_suspend) ;


	DBGTRACE("%s(%d)", __func__, __LINE__) ;

	if( p_priv->init ) {
		run(p_priv, NULL, 0) ;
	}
}


/* -------------------------------------------------------------------------------- */
/*   プローブ                                                                       */
/*                                                                                  */
/*   IN : p_client = I2C クライアント                                               */
/*        p_id     = I2C デバイスの識別子                                           */
/*   OUT: 0       == 正常終了                                                       */
/*        0       <  異常終了                                                       */
/* -------------------------------------------------------------------------------- */
static int mgc3x30_probe(struct i2c_client *p_client, const struct i2c_device_id *p_id)
{
	struct mgc3x30_config	*p_config=p_client->dev.platform_data ;
	struct priv				*p_priv ;
	u32						const *p_keyCode ;
	int						result ;
	uint					index ;

	/* プライベートデータ確保 */
	p_priv = kzalloc(sizeof( struct priv ), GFP_KERNEL) ;
	if( p_priv == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		return -ENOMEM ;
	}

	/* 初期化 */
	i2c_set_clientdata(p_client, p_priv) ;
	init_waitqueue_head( &p_priv->waitQueueHead ) ;
	p_priv->p_client = p_client ;
	p_priv->p_config = p_config ;
	p_priv->eint     = p_config->eint ;
	p_priv->mclr     = p_config->mclr ;
	p_priv->eio0     = p_config->eio0 ;

	/* MISC デバイス作成 */
	p_priv->misc.minor = MISC_DYNAMIC_MINOR ;
	p_priv->misc.name  = DRV_NAME ;
	p_priv->misc.fops  = &mgc3x30_fops ;
	result             = misc_register( &p_priv->misc ) ;
	if( result ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		result = -ENOMEM ;
		goto mgc3x30_probe_f00 ;
	}

	/* 入力デバイス確保 */
	p_priv->p_input = input_allocate_device() ;
	if( p_priv->p_input == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		result = -ENOMEM ;
		goto mgc3x30_probe_e00 ;
	}

	/* 入力デバイス登録 */
	p_keyCode = &p_config->p_key_code[0] ;
	for(index=0; index<p_config->count; index++,p_keyCode++) {
		__set_bit(*p_keyCode, p_priv->p_input->keybit) ;
	}
	__set_bit(EV_KEY, p_priv->p_input->evbit) ;
	p_priv->p_input->name       = DRV_NAME ;
	p_priv->p_input->phys       = DRV_NAME ;
	p_priv->p_input->id.bustype = BUS_HOST ;
	p_priv->p_input->id.vendor  = 0x0001 ;
	p_priv->p_input->id.product = 0x0001 ;
	p_priv->p_input->id.version = 0x0100 ;
	p_priv->p_input->dev.parent = &p_client->dev ;
	result                      = input_register_device( p_priv->p_input ) ;
	if( result ) {
		input_free_device( p_priv->p_input ) ;
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		goto mgc3x30_probe_e00 ;
	}

	/* DMA バッファ確保 */
	p_priv->p_buff = dma_alloc_coherent(0, BUFF_SIZE, &p_priv->buff_addr, GFP_KERNEL) ;
	if( p_priv->p_buff == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		result = -ENOMEM ;
		goto mgc3x30_probe_d00 ;
	}

	/* ワークキュー作成 */
	p_priv->p_workqueue = create_workqueue( DRV_NAME ) ;
	if( p_priv->p_workqueue == NULL ) {
		ERRTRACE("%s(%d)", __func__, __LINE__) ;
		result = -ENOMEM ;
		goto mgc3x30_probe_c00 ;
	}
	INIT_WORK(&p_priv->work, (void *)workqueue_proc) ;

	m_p_priv = p_priv ;

	/* サスペンド登録 */
	p_priv->early_suspend.level   = EARLY_SUSPEND_LEVEL_BLANK_SCREEN + 1 ;
	p_priv->early_suspend.suspend = mgc3x30_early_suspend ;
	p_priv->early_suspend.resume  = mgc3x30_early_resume ;
	register_early_suspend( &p_priv->early_suspend ) ;

	/* 実行 */
	result = run(p_priv, &Library, FLAG_CHECKVERSION) ;
	if( !result ) {
		p_priv->init = 1 ;
	}

	return 0 ;


mgc3x30_probe_c00:
	dma_free_coherent(0, BUFF_SIZE, &p_priv->p_buff[0], p_priv->buff_addr) ;
mgc3x30_probe_d00:
	input_unregister_device( p_priv->p_input ) ;
mgc3x30_probe_e00:
	misc_deregister( &p_priv->misc ) ;
mgc3x30_probe_f00:
	kfree( p_priv ) ;

	return result ;
}


/* -------------------------------------------------------------------------------- */
/*   削除                                                                           */
/*                                                                                  */
/*   IN : p_client = I2C クライアント                                               */
/*   OUT: 常に０                                                                    */
/* -------------------------------------------------------------------------------- */
static int mgc3x30_remove( struct i2c_client *p_client )
{
	struct priv		*p_priv=i2c_get_clientdata( p_client ) ;


	unregister_early_suspend( &p_priv->early_suspend ) ;

	stop( p_priv ) ;

	destroy_workqueue( p_priv->p_workqueue ) ;

	dma_free_coherent(0, BUFF_SIZE, &p_priv->p_buff[0], p_priv->buff_addr) ;

	input_unregister_device( p_priv->p_input ) ;

	misc_deregister( &p_priv->misc ) ;

	kfree( p_priv ) ;

	return 0 ;
}


static const struct i2c_device_id	mgc3x30_id[] = {
										{ "mgc3030", 3030 },
										{ "mgc3130", 3130 },
										{}
									} ;

static struct i2c_driver			mgc3x30_driver = {
										.probe     = mgc3x30_probe,
										.remove    = mgc3x30_remove,
										.id_table  = mgc3x30_id,
										.driver    = {
											.name  = DRV_NAME,
											.owner = THIS_MODULE,
										},
									} ;


module_i2c_driver( mgc3x30_driver ) ;


MODULE_DESCRIPTION("Sony MGC3x30 Driver");
MODULE_AUTHOR("Sony Corporation");
MODULE_LICENSE("GPL") ;
MODULE_ALIAS("Sony MGC3x30 Driver") ;
MODULE_VERSION( "v0.01" ) ;
